/********************************************************/
/* Copyright (C) 2001 OSS Nokalva.  All rights reserved.*/
/********************************************************/

/* THIS FILE IS PROPRIETARY MATERIAL OF OSS Nokalva, INC. AND
 * MAY BE USED ONLY BY DIRECT LICENSEES OF OSS Nokalva, INC.
 * THIS FILE MAY NOT BE DISTRIBUTED. */

/* Prepared by OSS Nokalva, Inc.  11/22/2000      */

/**
    Application program MyBCD.java (part of Usebcd.java)
    Demonstrates the UserClass directive of the OSS ASN.1/Pure Java Tools
    API using bcd.asn and MyBCD.java. To run this program:

    asn1pjav bcd.asn
    move MyBCD.java bcd
    cd bcd
    bcd.bat javac
    cd ..
    javac -g Usebcd.java
    java Usebcd

    MyBCD.java extends the generated class that had the UserClass directive.
    It implements the Printable interface as an illustration.
    It is used as a replacement for the compiler-generated class BCDString.
    MyBCD allows the encoded version of the value to be in binary coded
    decimal (BCD) format, while allowing the value at any time to be printable
    in ascii (readable) format.

 */

package bcd;

import com.oss.asn1.*;
import com.oss.metadata.*;
import bcd.a.BCDString;
import java.io.PrintWriter;

/**
 * Illustrates the com.oss.asn1.Printable interface by implementing the
 * print(int, PrintWriter) method.
 *
 * Contains conversion methods between BCD and ascii format: toBCD() and fromBCD().
 * @see OctetString
 */

public class MyBCD extends BCDString implements Printable {

    /**
     * The default constructor.
     */
    public MyBCD()
    {
    }

    /**
     * Construct from a byte[] type.
     * @param the byte[] to set this object to.
     */

    public MyBCD(byte[] value)
    {
	super(value);
    }

    /**
     * Construct BCD from the decimal number
     * @exception java.lang.NumberFormatException if BCD conversion error
     * occurs
     */
    public MyBCD(String number) throws NumberFormatException
    {
	super(MyBCD.toBCD(number));
    }

    /**
     * Construct BCD from the binary number (long)
     * @exception java.lang.NumberFormatException if BCD conversion error
     * occurs
     */
    public MyBCD(long number) throws NumberFormatException
    {
	super(MyBCD.toBCD(number));
    }

    // Converts binary number to its BCD encoding.
    public static byte[] toBCD(long number) throws NumberFormatException
    {
	return toBCD(Long.toString(number));
    }

    // Converts decimal number to its BCD encoding.
    public static byte[] toBCD(String number)
	throws NumberFormatException
    {
	// Number of digits in 'number'
	int length = number.length();

	if (length < 1)
	    throw new NumberFormatException("no decimal digits");

	// Allocate byte[] array to store BCD digits
	byte[] result = new byte[(length+1)/2];

	// If the 'number' contains odd number of digits, the rightmost
	// nibble in the last octet will be padded with 0xF
	boolean padded = (length % 2) == 1;
	if (padded)
	    --length;

	// Pack decimal digits (two digits come to one octet)
	int decimalPos, bcdPos;
	for (decimalPos=0, bcdPos=0; decimalPos<length; decimalPos++, bcdPos++)
	    result[bcdPos] = (byte)((digit(number.charAt(decimalPos)) << 4) |
		digit(number.charAt(++decimalPos)));

	// If the number of decimal digits is odd, send the remaining decimal
	// digit and fill the rightmost nibble of the last octet with 0xF
	if (padded)
	    result[bcdPos] = (byte)((digit(number.charAt(decimalPos)) << 4) |
		0x0F);

	return result;
    }

    // Returns the numeric value of the character 'decimalDigit' for the
    // radix of 10.
    private static int digit(char decimalDigit) throws NumberFormatException
    {
	int d = Character.digit(decimalDigit, 10);
	if (d == -1)
	    // Invalid decimal digit. Throw an exception.
	    throw new NumberFormatException(
		"bad decimal digit '" + decimalDigit + "'");

	return d;
    }

    // Converts BCD encoding to decimal number
    public static String fromBCD(byte[] number) throws NumberFormatException
    {
	// Number of BCD digits
	int length = number.length;

	if (length < 1)
	    throw new NumberFormatException("no BCD digits");

	// Check if the least significant nibble is a padding nibble
	boolean padded = (number[length-1] & 0x0F) == 0x0F;
	if (padded)
	    --length;

	// Compute the number of decimal digits in the BCD
	int size = 2* length;
	if (padded)
	    ++size;

	// Allocate StringBuffer to receive converted decimal number
	StringBuffer decimalNumber = new StringBuffer(size);

	int bcdPos;

	// Extract decimal digits from the BCD encoding and store them
	// into the buffer
	for (bcdPos=0; bcdPos<length; bcdPos++) {
	    decimalNumber.append(forDigit((number[bcdPos] >> 4) & 0x0F));
	    decimalNumber.append(forDigit((number[bcdPos]) & 0x0F));
	}

	// BCD contained odd number of BCD digit. Extract the remaining
	// digit and ignore the trailing nibble with 0xF
	if (padded)
	    decimalNumber.append(forDigit((number[bcdPos] >> 4) & 0x0F));

	return decimalNumber.toString();
    }

    // Returns the character that represents decimal digit 'd'.
    private static char forDigit(int d) {
	char c = Character.forDigit(d, 10);

	if (c == 0)
	    // Bad decimal digit (<0 or >9). Throw an exception.
	    throw new NumberFormatException("bad BCD digit " + d);

	return c;
    }


    /**
     * Implement the print(int, PrintWriter) method of the Printable
     * interface. Special care should be taken if the code needs to call the
     * toString() method of the AbstractData in order to get the defailt value
     * notation. Since the toString() implicitly invokes the print() method,
     * it might result in infinite recursion.
     */
    public int print(int mode, PrintWriter s) throws UserPrintFailedException
    {
	if (mode == Printable.BEFORE) {
	    try {
		String bcdNumber = fromBCD(byteArrayValue());
		s.print(bcdNumber);
		return Printable.DONE;
	    } catch (NumberFormatException e) {
		// Report that the value has invalid format and
		// gracefully print the raw contents
		s.print(" -- " + e.getMessage() + " -- ");
		return Printable.CONTINUE;
	    } catch (Exception e) {
		// Throw an UserPrintFailedException in case of
		// serious error
		throw new UserPrintFailedException(e.toString());
	    }
	} else  // mode is Printable.AFTER
	    return Printable.DONE;
    }

} // End class definition for MyBCD
